@@ -10,11 +10,14 @@ from pathlib import Path
 from typing import (
     Callable,
     Dict,
+    Iterable,
+    Iterator,
     List,
     NamedTuple,
     Optional,
     Sequence,
     Tuple,
+    TYPE_CHECKING,
     TypeVar,
     Union,
 )
@@ -26,7 +29,11 @@ from flake8.main.application import (  # type: ignore[import]
 from flake8.style_guide import Decision  # type: ignore[import]
 from parso.python import tree
 
-FIXER_REGEX = re.compile(r'^fix_([A-Z]\d{3})$')
+if TYPE_CHECKING:
+    from typing_extensions import TypeGuard
+
+
+FIXER_REGEX = re.compile(r'^fix_([A-Z]{1,3}\d{3})$')
 
 FLAKE8_FORMAT = '%(path)s:%(row)d:%(col)d:%(code)s:%(text)s'
 
@@ -37,6 +44,9 @@ class ErrorDetail(NamedTuple):
     code: str
     message: str
 
+    def render(self, filename: Path) -> str:
+        return f'{filename}:{self.line}:{self.col}: {self.code} {self.message}'
+
 
 class CodeLine(NamedTuple):
     text: str
@@ -49,10 +59,14 @@ Span = Tuple[Position, Position]
 
 # TODO: currently we return the new string, we should move to instead returning
 # a description of the edit.
-Fixer = Callable[[CodeLine], str]
-TFixer = TypeVar('TFixer', bound=Fixer)
+LineFixer = Callable[[CodeLine], str]
+TLineFixer = TypeVar('TLineFixer', bound=LineFixer)
 
-FIXERS: Dict[str, Fixer] = {}
+FileFixer = Callable[[Sequence[ErrorDetail], str], str]
+TFileFixer = TypeVar('TFileFixer', bound=FileFixer)
+
+LINE_FIXERS: Dict[str, LineFixer] = {}
+FILE_FIXERS: Dict[str, FileFixer] = {}
 
 
 def merge_overlapping_spans(spans: Sequence[Span]) -> List[Span]:
@@ -86,48 +100,60 @@ def remove_character_at(text: str, col: int, char: str) -> str:
     return text[:col] + text[col + 1:]
 
 
-def fixer(fn: TFixer) -> TFixer:
+def line_fixer(fn: TLineFixer) -> TLineFixer:
     match = FIXER_REGEX.match(fn.__name__)
     if match is None:
         raise ValueError(
-            "Fixer has invalid name, should be of the form 'fix_X123' but was "
+            "LineFixer has invalid name, should be of the form 'fix_X123' but was "
             "{!r}".format(fn.__name__),
         )
-    FIXERS[match.group(1)] = fn
+    LINE_FIXERS[match.group(1)] = fn
     return fn
 
 
-@fixer  # Missing trailing comma
+def file_fixer(fn: TFileFixer) -> TFileFixer:
+    match = FIXER_REGEX.match(fn.__name__)
+    if match is None:
+        raise ValueError(
+            "LineFixer has invalid name, should be of the form 'fix_X123' but was "
+            "{!r}".format(fn.__name__),
+        )
+    FILE_FIXERS[match.group(1)] = fn
+    return fn
+
+
+@line_fixer  # Missing trailing comma
 def fix_C812(code_line: CodeLine) -> str:
     return insert_character_at(code_line.text, code_line.col, ',')
 
 
-@fixer  # Missing trailing comma
+@line_fixer  # Missing trailing comma
 def fix_C813(code_line: CodeLine) -> str:
     return insert_character_at(code_line.text, code_line.col, ',')
 
 
-@fixer  # Missing trailing comma
+@line_fixer  # Missing trailing comma
 def fix_C814(code_line: CodeLine) -> str:
     return insert_character_at(code_line.text, code_line.col, ',')
 
 
-@fixer  # Missing trailing comma
+@line_fixer  # Missing trailing comma
 def fix_C815(code_line: CodeLine) -> str:
     return insert_character_at(code_line.text, code_line.col, ',')
 
 
-@fixer  # Missing trailing comma
+@line_fixer  # Missing trailing comma
 def fix_C816(code_line: CodeLine) -> str:
     return insert_character_at(code_line.text, code_line.col, ',')
 
 
-@fixer  # Trailing comma prohibited
+@line_fixer  # Trailing comma prohibited
 def fix_C819(code_line: CodeLine) -> str:
     # flake8-commas seems to give the wrong column position, so -1
     return remove_character_at(code_line.text, code_line.col - 1, ',')
 
 
+@file_fixer
 def fix_F401(messages: Sequence[ErrorDetail], content: str) -> str:
     module = parso.parse(content).get_root_node()
 
@@ -195,7 +221,7 @@ def fix_F401(messages: Sequence[ErrorDetail], content: str) -> str:
         name: tree.Name,
         import_as_name: Optional[str],
         node: Union[tree.ImportFrom, tree.ImportName],
-    ) -> parso.tree.NodeOrLeaf:
+    ) -> Union[parso.tree.BaseNode, parso.tree.Leaf]:
         assert name.parent is not None  # placate mypy
         if name.parent.parent == node:
             # We're removing something like `bar as spam` from
@@ -208,47 +234,56 @@ def fix_F401(messages: Sequence[ErrorDetail], content: str) -> str:
         assert import_as_name, "Did not expect renamed import, but found one"
         return name.parent
 
+    def is_operator(node: parso.tree.NodeOrLeaf, char: str) -> 'TypeGuard[tree.Operator]':
+        return isinstance(node, tree.Operator) and node == char
+
+    def operators_to_remove(
+        elements: Iterable[parso.tree.NodeOrLeaf],
+    ) -> Iterator[tree.Operator]:
+        last_was_operator = True
+        for element in elements:
+            if not is_operator(element, ','):
+                last_was_operator = False
+                continue
+
+            if last_was_operator:
+                yield element
+
+            last_was_operator = True
+
     def get_parts_to_remove(
         line_messages: Sequence[ErrorDetail],
         node: Union[tree.ImportFrom, tree.ImportName],
     ) -> List[Span]:
-        spans_to_remove: List[Span] = []
+        nodes_to_remove = set(
+            get_node_to_remove(
+                *get_part_to_remove(message, node),
+                node,
+            )
+            for message in line_messages
+        )
 
-        last_comma_was_trailing = False
-        last_end_pos = (0, 0)
+        parent, = set(x.parent for x in nodes_to_remove)
+        assert parent is not None  # placate mypy
+        remaining_siblings = [x for x in parent.children if x not in nodes_to_remove]
 
-        for message in line_messages:
-            last_part, import_as_name = get_part_to_remove(message, node)
-            node_to_remove = get_node_to_remove(last_part, import_as_name, node)
+        nodes_to_remove.update(operators_to_remove(remaining_siblings))
 
-            start_pos = get_start_pos(node_to_remove)
-            end_pos = node_to_remove.end_pos
+        last_remaining_sibling, = itertools.islice(
+            (x for x in reversed(remaining_siblings) if x not in nodes_to_remove),
+            1,
+        )
 
-            comma_is_trailing = False
+        if (
+            is_operator(last_remaining_sibling, ',')
+            and not is_operator(node.children[-1], ')')
+        ):
+            nodes_to_remove.add(last_remaining_sibling)
 
-            # Look for commas to remove. We prefer to remove a preceding comma,
-            # will pick up a trailing one if that's going to work better.
-            prev_leaf = last_part.get_previous_leaf()
-            if (
-                # Check that the preceding comma is on the same line.
-                on_same_line(prev_leaf, last_part)
-                and prev_leaf.type == 'operator'
-                # Check that that preceding comma isn't already being removed.
-                and not (last_comma_was_trailing and start_pos <= last_end_pos)
-            ):
-                start_pos = get_start_pos(prev_leaf)
-
-            else:
-                next_leaf = node_to_remove.get_next_leaf()
-                if next_leaf.type == 'operator':
-                    end_pos = next_leaf.end_pos
-                    comma_is_trailing = True
-
-            last_comma_was_trailing = comma_is_trailing
-            last_end_pos = end_pos
-            spans_to_remove.append((start_pos, end_pos))
-
-        return spans_to_remove
+        return [
+            (get_start_pos(x), x.end_pos)
+            for x in nodes_to_remove
+        ]
 
     spans_to_remove: List[Span] = []
 
@@ -286,11 +321,25 @@ def fix_F401(messages: Sequence[ErrorDetail], content: str) -> str:
 
         interim = [] if line.isspace() else [line]
 
+        if not interim and not before:
+            # We're removing a line at the start of the file. If the line after
+            # is also empty, then we also want to remove those lines to avoid
+            # creating empty lines at the start of the file. This is a specific
+            # targetted fix rather than a general attempt at formatting imports
+            # (we leave that to e.g: `isort`).
+            if after and after[0].isspace():
+                after.pop(0)
+
         lines = before + interim + after
 
     return ''.join(lines)
 
 
+@file_fixer
+def fix_LBL001(messages: Sequence[ErrorDetail], content: str) -> str:
+    return content.lstrip('\n')
+
+
 def parse_flake8_output(flake8_output: str) -> Dict[Path, List[ErrorDetail]]:
     """
     Parse output from Flake8 formatted using FLAKE8_FORMAT into a useful form.
@@ -324,7 +373,7 @@ def run_flake8(args: List[str]) -> Dict[Path, List[ErrorDetail]]:
         decider = flake8.guide.decider
         flake8.options.select = [
             code
-            for code in FIXERS.keys()
+            for code in LINE_FIXERS.keys()
             if decider.decision_for(code) == Decision.Selected
         ]
         flake8.run_checks()
@@ -345,14 +394,14 @@ def process_errors(messages: List[ErrorDetail], content: str) -> str:
     modified = False
 
     for message in sorted(messages, reverse=True):
-        fixer_fn = FIXERS.get(message.code)
-        if not fixer_fn:
+        line_fixer_fn = LINE_FIXERS.get(message.code)
+        if not line_fixer_fn:
             continue
 
         # Convert to 0-based
         lineno = message.line - 1
 
-        new_line = fixer_fn(CodeLine(lines[lineno], lineno, message.col))
+        new_line = line_fixer_fn(CodeLine(lines[lineno], lineno, message.col))
         if new_line == lines[lineno]:
             continue
 
@@ -362,10 +411,10 @@ def process_errors(messages: List[ErrorDetail], content: str) -> str:
     if modified:
         content = ''.join(x.rstrip() + '\n' for x in lines)
 
-    # TODO: generalise support for other whole-file fixes
-    f401_messages = [x for x in messages if x.code == 'F401']
-    if f401_messages:
-        content = fix_F401(f401_messages, content)
+    for code, fixer_fn in FILE_FIXERS.items():
+        relevant_messages = [x for x in messages if x.code == code]
+        if relevant_messages:
+            content = fixer_fn(relevant_messages, content)
 
     return content
 
@@ -374,6 +423,10 @@ def run(args: argparse.Namespace) -> None:
     all_error_details = run_flake8(args.flake8_args)
 
     for filepath, error_details in all_error_details.items():
+        if error_details[0].code == 'E999':
+            print(error_details[0].render(filepath))
+            continue
+
         with filepath.open(mode='r+') as f:
             content = f.read()
             new_content = process_errors(error_details, content)
